
#include "svfparser.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <deque>
#include <list>
#include <map>
#include <exception>
#include <stdexcept>
#include <algorithm>

#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string/erase.hpp>
#include <boost/algorithm/string/find.hpp>
//#include <boost/algorithm/string/join.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/regex.hpp>
#include <boost/algorithm/string/replace.hpp>

//#include <boost/assign.hpp>
#include <boost/assign/list_inserter.hpp>
#include <boost/assign/std/list.hpp>
#include <boost/assign/std/vector.hpp>

#include "colorlog/ansi_col.h"
#include "mlink/stl_ext.h"
#include "mlink/exc.h"
#include "mlink/jtag_state.h"
#include "BitVect.h"

namespace mlink {

    using namespace boost::algorithm;
    //using namespace boost::assign;

    template <class T>
            void printIt(T &r)
    {
        std::cout << r << std::endl;
    }

    std::ostream &operator<< (std::ostream & os, const JtagShift &r)
    {
        os << "JtagShift[" << r.size() << "]";
        if (r.size() > 50) {
            os << ", tdi={...}, tdo={...}";
            return os;
        }

        /*
  os << ", tms={";
  for(JtagShift::const_iterator it = r.begin(); it != r.end(); ++it) {
        os << (*it).tms;
  }
  os << "}";
*/
        os << ", tdi={";
        for(JtagShift::const_iterator it = r.begin(); it != r.end(); ++it) {
            os << (*it).tdi;
        }
        os << "}";

        os << ", tdo={";
        for(JtagShift::const_iterator it = r.begin(); it != r.end(); ++it) {
            os << (*it).tdo;
        }
        os << "}";

        return os;
    }


    std::ostream& operator<< (std::ostream & os, const TokenList &r)
    {
        os << "tokens={";
        if (r.size() > 0) {
            os << r.front();
            for(TokenList::const_iterator it = ++r.begin(); it!=r.end(); ++it) {
                os << "," << (*it);
            }
        }
        os << "}";
        return os;
    }

    std::ostream& operator<< (std::ostream & os, const SvfOp &r)
    {
        os << "Op: " << r.name;
        return os;
    }

    std::ostream& operator<< (std::ostream & os, const SvfOpSxr &r)
    {
        os << "OpSxr: " << r.name;
        return os;
    }

    std::ostream& operator<< (std::ostream & os, const SvfList &r)
    {
        os << "Commands:" << std::endl;
        for(SvfList::const_iterator it = r.begin(); it != r.end(); ++it) {
            os << *(*it) << std::endl;
        }
        os << std::endl;
        return os;
    }

    class SvfTapState: public std::map<std::string, JTAG_STATE> {
    public:
        SvfTapState();
        JTAG_STATE getState(const std::string &) const;
    };

    SvfTapState::SvfTapState()
    {
        boost::assign::insert(*this)
                ("RESET", JS_TEST_LOGIC_RESET)
                ("IDLE", JS_RUN_TEST_IDLE)
                ("DRSELECT", JS_SELECT_DR_SCAN)
                ("DRCAPTURE", JS_CAPTURE_DR)
                ("DRSHIFT", JS_SHIFT_DR)
                ("DREXIT1", JS_EXIT1_DR)
                ("DRPAUSE", JS_PAUSE_DR)
                ("DREXIT2", JS_EXIT2_DR)
                ("DRUPDATE", JS_UPDATE_DR)
                ("IRSELECT", JS_SELECT_IR_SCAN)
                ("IRCAPTURE", JS_CAPTURE_IR)
                ("IRSHIFT", JS_SHIFT_IR)
                ("IREXIT1", JS_EXIT1_IR)
                ("IRPAUSE", JS_PAUSE_IR)
                ("IREXIT2", JS_EXIT2_IR)
                ("IRUPDATE", JS_UPDATE_IR);
    }

    JTAG_STATE SvfTapState::getState(const std::string &s) const
    {
        SvfTapState::const_iterator it = SvfTapState::find(s);
        if (it == end()) throw std::runtime_error("invalid TAP state "+s);
        return (*it).second;
    }

    std::ostream& operator<< (std::ostream & os, const SvfShiftOptions &r)
    {
        os << "tdi[" << r.tdi.size() << "], tdo[" << r.tdo.size() << "]";
        return os;
    }

    JtagShift::JtagShift(const SvfShiftOptions &opt)
    {
        const JtagBits &tdi = opt.tdi;
        const JtagBits &mdi = opt.smask;
        const JtagBits &tdo = opt.tdo;
        const JtagBits &mdo = opt.mask;

        const bool has_tdi = !tdi.empty();
        const bool has_tdo = !tdo.empty();
        const bool has_mdi = !mdi.empty();
        const bool has_mdo = !mdo.empty();

        if (has_tdi && has_tdo && (tdo.size() != tdi.size())) {
            std::ostringstream os;
            os << "size mismatch: tdi[" << tdi.size() << "], tdo[" << tdo.size() << "]";
            throw std::runtime_error(os.str());
        }
        if (has_tdi && has_mdi && (tdi.size() != mdi.size())) {
            std::ostringstream os;
            os << "size mismatch: tdi[" << tdi.size() << "], smask[" << mdi.size() << "]";
            throw std::runtime_error(os.str());
        }
        if (has_tdo && has_mdo && (tdo.size() != mdo.size())) {
            std::ostringstream os;
            os << "size mismatch: tdo[" << tdo.size() << "], mask[" << mdo.size() << "]";
            throw std::runtime_error(os.str());
        }
        size_t n = std::max(tdi.size(), tdo.size());
        if (n==0) {
            std::ostringstream os;
            os << "TDI: " << tdi.size() << std::endl;
            os << "MDI: " << mdi.size() << std::endl;
            os << "TDO: " << tdo.size() << std::endl;
            os << "MDO: " << mdo.size() << std::endl;
            std::cerr << os.str();
            throw std::runtime_error("JtagShift: empty");
        }
        resize(n);
        if (opt.endState != opt.shiftState) {
            (*this)[n-1].tms = 1;
        }

        if (has_tdi) {
            for(size_t i=0; i<opt.tdi.size(); i++) {
                bool smask = has_mdi?opt.smask[i]:1;
                ((*this)[i]).tdi = opt.tdi[i];
                ((*this)[i]).smask = smask;
            }
        }
        if (has_tdo) {
            for(size_t i=0; i<opt.tdo.size(); i++) {
                bool mask = has_mdo?opt.mask[i]:1;
                ((*this)[i]).tdo = opt.tdo[i];
                ((*this)[i]).mask = mask;
            }
        }
    }

    JtagShift SvfOpState::getSequence() const
    {
        return JtagShift(); // FIXME
    }


    SvfParser::SvfParser(const std::string &fileName)
    {
        frequency = 4.0e6;
        svfTapState = new SvfTapState();
        createKnownTokenList();

        std::ifstream ifs;
        ifs.open(fileName.c_str(), std::ifstream::in);
        if (!ifs.good()) {
            std::ostringstream os;
            os << "Can't open file '" << fileName << "'";
            throw std::runtime_error(os.str());
        }
        std::string line;
        while(ifs.good()) {
            std::string origline;
            getline(ifs, origline);
            replace_all(origline, "\n\r", "\n");
            replace_all(origline, "\r\n", "\n");
            replace_all(origline, "\r", "\n");
            try {
                //	trim(line);
                if (origline.substr(0, 2) == "//") {
                    std::string comment = origline;
                    erase_all(comment, "\n");
                    //	  std::cout << comment << endl;
                    boost::shared_ptr<SvfOpComment> p(new SvfOpComment(comment));
                    commands.push_back(p);
                    continue;
                }
                //	erase_all(line, "\n");
                //	erase_all(line, "\r");
                line += origline;
                bool origline_terminated = origline.find(";") != std::string::npos;
                while(origline_terminated && !line.empty()) {
                    size_t found = line.find(";");
                    if (found == std::string::npos) break;
                    std::string subline = line.substr(0, found);
                    erase_all(subline, "\n");
                    trim(subline);
                    //	  cerr << "subl: '" << subline << "'" << endl;

                    /*
	  /// print original command as a comment ///
          boost::shared_ptr<SvfOpComment> p(
                  new SvfOpComment(std::string(ANSI_BLUE) + std::string("// '")
                                   + subline
                                   + std::string("'") + std::string(ANSI_NORM)));
          commands.push_back(p);
*/
                    line.erase(0, found+1);
                    //	  cerr << "line: '" << line << "'" << endl;

                    std::vector<std::string> tokens;
                    split(tokens, subline, is_space() );
                    if (!tokens.empty() && !tokens[0].empty()) {
                        parseTokens(tokens);
                    }

                    //	  usleep(100000);
                }
            }
            catch(std::runtime_error &e) {
                std::cerr << "Error parsing line: " << origline << std::endl;
                throw;
            }
        }
        ifs.close();
    }

    SvfParser::~SvfParser()
    {
        //  printIt(commands);
        delete svfTapState;
    }

    JTAG_STATE SvfParser::getSvfTapState(const std::string &str) const
    {
        return svfTapState->getState(str);
    }

    struct StringCompare: std::binary_function<std::string, std::string, bool> {
        bool operator()(const std::string &lhs, const std::string &rhs) const { return lhs == rhs; }
    };

    bool TokenList::contains(const std::string &name) const
    {
        return find_if(begin(), end(), std::bind2nd(StringCompare(), name) ) != end();
    }

    void SvfParser::checkCountEq(const TokenList &t, size_t count) const
    {
        if (t.size() != count)
            throw svf_parse_error(": invalid number of arguments");
    }

    void SvfParser::checkCountGe(const TokenList &t, size_t count) const
    {
        if (t.size() < count)
            throw svf_parse_error(": insufficient number of arguments");
    }

    void SvfParser::createKnownTokenList()
    {
        using namespace boost::assign;
        endXrTokens += "ENDDR", "ENDIR";
        xxrTokens += "HDR", "HIR", "TDR", "TIR", "SDR", "SIR";
    }

    void SvfParser::parseTokens(const std::vector<std::string> &tokens)
    {
        TokenList t;
        std::copy(tokens.begin(), tokens.end(), back_inserter(t));
        if (t.empty() || t.front().empty()) return;

        //  printIt(t);
        const std::string name = t.front();
        t.pop_front();
        //	std::cout << name << std::endl;

        if (name == "FREQUENCY") {
            //	cout << "Freq" << endl;
            checkCountGe(t, 1);
            /*
/// Disabled parsing of FREQUENCY token = use fixed TCK frequency 12.5 MHz ///
    double freqHz = strtod(t.front().c_str(), 0);
    frequency = freqHz;
*/
            return;
        }
        if (name == "TRST") {
            //	cout << "Freq" << endl;
            checkCountGe(t, 1);
            // FIXME
            return;
        }
        if (name == "STATE") {
            //	cout << "State" << endl;
            checkCountGe(t, 1);
            JTAG_STATE state = getSvfTapState(t.front());
            boost::shared_ptr<SvfOpState> p(new SvfOpState(state));
            commands.push_back(p);
            return;
        }
        if (name == "RUNTEST") {
            //	cout << "State" << endl;
            checkCountGe(t, 2);

            unsigned int count = boost::lexical_cast<unsigned int>(t.front());
            t.pop_front();
            const std::string opname = t.front();
            if (opname != "TCK")
                throw svf_parse_error("Unknown RUNTEST argument 2, expected TCK");
            double delay = count/frequency;
            boost::shared_ptr<SvfOpRunTest> p(new SvfOpRunTest(delay));
            commands.push_back(p);
            return;
        }
        if (endXrTokens.contains(name)) {
            //	cout << "EndXr" << endl;
            parseTokensEndXr(name, t);
            return;
        }

        if (xxrTokens.contains(name)) {
            //	cout << "XXr" << endl;
            parseTokensXxr(name, t);
            return;
        }
        std::ostringstream os;
        os << "Unknown token '" << name << "'" << std::endl;
        throw std::runtime_error(os.str());
    }

    void SvfParser::parseTokensEndXr(const std::string &name, TokenList &t)
    {
        checkCountGe(t, 1);
        JTAG_STATE endState = getSvfTapState(t.front());
        if (name == "ENDDR")
            endDr = endState;
        if (name == "ENDIR")
            endIr = endState;
    }

    void SvfParser::parseTokensXxr(const std::string &name, TokenList &t)
    {
        checkCountGe(t, 1);
        unsigned int bitCount = boost::lexical_cast<unsigned int>(t.front());
        t.pop_front();
        JtagBits tdi;
        //  JtagBits smask;
        JtagBits tdo;
        //  JtagBits mask;
        bool is_sxr = (name == "SDR") || (name == "SIR");
        if (name == "HDR") hxr["SDR"].clear();
        if (name == "TDR") txr["SDR"].clear();
        if (name == "HIR") hxr["SIR"].clear();
        if (name == "TIR") txr["SIR"].clear();
        //  cerr << "name: " << name << endl;
        while (!t.empty()) {
            //	t.print();
            checkCountGe(t, 2);
            const std::string tag = t.front();
            t.pop_front();
            const std::string value = t.front();
            t.pop_front();
            std::string subval;
            if ((value.size() >= 2) && (value.substr(0, 1) == "(") && (value.substr(value.size()-1, 1) == ")")) {
                subval = value.substr(1, value.size()-2); // trim first and last characters
            } else {
                std::ostringstream os;
                os << "tag '" << tag << "', expected '(...)', got '" << value << "'";
                throw std::runtime_error(os.str());
            }

            //	cerr << "tag: " << tag << endl;
            BitVect bv = BitVect(bitCount, subval);
            //	cerr << "tag: " << tag << " " << bv << endl;
            JtagBits bits = bv; // parse hex string to bits
            /*
        std::map<std::string, BitVect> header;
        std::map<std::string, BitVect> trailer;
        if (name == "SDR") {
          header = hdr;
          trailer = tdr;
        }
        if (name == "SIR") {
          header = hir;
          trailer = tir;
        }
*/
            /*
        if (is_sxr) {
          if (!header ["TDI"].empty() && header ["TDO"].empty()) { cout << "copy H TDI->TDO" << endl; header ["TDO"] = header ["TDI"]; }
          if (!trailer["TDI"].empty() && trailer["TDO"].empty()) { cout << "copy T TDI->TDO" << endl; trailer["TDO"] = trailer["TDI"]; }

          if (!header ["TDI"].empty() && header ["SMASK"].empty()) { cout << "copy H TDI->SMASK" << endl; header ["SMASK"] = header ["TDI"]; }
          if (!trailer["TDI"].empty() && trailer["SMASK"].empty()) { cout << "copy T TDI->SMASK" << endl; trailer["SMASK"] = trailer["TDI"]; }

          if (!header ["TDO"].empty() && header ["MASK"].empty()) { cout << "copy H TDO->MASK" << endl; header ["MASK"] = header ["TDO"]; }
          if (!trailer["TDO"].empty() && trailer["MASK"].empty()) { cout << "copy T TDO->MASK" << endl; trailer["MASK"] = trailer["TDO"]; }
        }
*/
            if (tag == "TDI") {
                tdi = bits;
            }
            if (tag == "SMASK") {
                save_smask[name] = bits;
            }
            if (tag == "TDO") {
                tdo = bits;
            }
            if (is_sxr && tag == "MASK") {
                save_mask[name] = bits;
            }
            if (name == "HDR") {
                hxr["SDR"][tag] = bits;
                //	  cout << "set HDR["<<tag<<"]=" << bits << endl;
            }
            if (name == "TDR") {
                txr["SDR"][tag] = bits;
                //	  cout << "set TDR["<<tag<<"]=" << bits << endl;
            }
            if (name == "HIR") {
                hxr["SIR"][tag] = bits;
                //	  cout << "set HIR["<<tag<<"]=" << bits << endl;
            }
            if (name == "TIR") {
                txr["SIR"][tag] = bits;
                //	  cout << "set TIR["<<tag<<"]=" << bits << endl;
            }
        }

        //  if (is_sxr) tdi.push_back(header[tag]);
        //	  tdi.push_back(bits);
        //  if (is_sxr) tdi.push_back(trailer[tag]);

        if (name == "SDR" || name == "SIR") {
            JtagBits h_tdi = hxr[name]["TDI"];
            JtagBits t_tdi = txr[name]["TDI"];
            JtagBits h_mdi = hxr[name]["SMASK"];
            JtagBits t_mdi = txr[name]["SMASK"];
            JtagBits h_tdo = hxr[name]["TDO"];
            JtagBits t_tdo = txr[name]["TDO"];
            JtagBits h_mdo = hxr[name]["MASK"];
            JtagBits t_mdo = txr[name]["MASK"];

            JtagBits mdi = save_smask[name];
            JtagBits mdo = save_mask[name];

            if (h_tdo.empty()) h_tdo.resize(h_tdi.size(), 0);
            if (t_tdo.empty()) t_tdo.resize(t_tdi.size(), 0);
            if (tdo.empty()) {tdo.resize(tdi.size(), 0); mdo.clear(); mdo.resize(tdi.size(), 0);}

            if (h_mdo.empty()) h_mdo.resize(h_mdi.size(), 0);
            if (t_mdo.empty()) t_mdo.resize(t_mdi.size(), 0);


            //    using namespace boost::assign;
            JtagBits alltdi;
            alltdi.reserve(h_tdi.size()+tdi.size()+t_tdi.size());
            alltdi.insert(alltdi.end(), h_tdi.begin(), h_tdi.end());
            alltdi.insert(alltdi.end(),   tdi.begin(),   tdi.end());
            alltdi.insert(alltdi.end(), t_tdi.begin(), t_tdi.end());

            JtagBits alltdo;
            alltdo.reserve(h_tdo.size()+tdo.size()+t_tdo.size());
            alltdo.insert(alltdo.end(), h_tdo.begin(), h_tdo.end());
            alltdo.insert(alltdo.end(),   tdo.begin(),   tdo.end());
            alltdo.insert(alltdo.end(), t_tdo.begin(), t_tdo.end());

            JtagBits allmdi;
            allmdi.reserve(h_mdi.size()+mdi.size()+t_mdi.size());
            allmdi.insert(allmdi.end(), h_mdi.begin(), h_mdi.end());
            allmdi.insert(allmdi.end(),   mdi.begin(),   mdi.end());
            allmdi.insert(allmdi.end(), t_mdi.begin(), t_mdi.end());

            JtagBits allmdo;
            allmdo.reserve(h_mdo.size()+mdo.size()+t_mdo.size());
            allmdo.insert(allmdo.end(), h_mdo.begin(), h_mdo.end());
            allmdo.insert(allmdo.end(),   mdo.begin(),   mdo.end());
            allmdo.insert(allmdo.end(), t_mdo.begin(), t_mdo.end());

            /*
        cout << endl;
        cout << "h_tdi: " <<  h_tdi << endl;
        cout << "  tdi: " <<    tdi << endl;
        cout << "t_tdi: " <<  t_tdi << endl;
        cout << "    =  " << alltdi << endl;

        cout << endl;
        cout << "h_mdi: " <<  h_mdi << endl;
        cout << "  mdi: " <<    mdi << endl;
        cout << "t_mdi: " <<  t_mdi << endl;
        cout << "    =  " << allmdi << endl;

        cout << endl;
        cout << "h_tdo: " <<  h_tdo << endl;
        cout << "  tdo: " <<    tdo << endl;
        cout << "t_tdo: " <<  t_tdo << endl;
        cout << "    =  " << alltdo << endl;

        cout << endl;
        cout << "h_mdo: " <<  h_mdo << endl;
        cout << "  mdo: " <<    mdo << endl;
        cout << "t_mdo: " <<  t_mdo << endl;
        cout << "    =  " << allmdo << endl;
*/
            //	SvfOpSdr op = SvfOpSdr(opt);
            //	printIt(op);
            if (name == "SDR") {
                SvfShiftOptions opt(alltdi, allmdi, alltdo, allmdo, JS_SHIFT_DR, endDr);
                boost::shared_ptr<SvfOpSdr> p(new SvfOpSdr(opt));
                commands.push_back(p);
            } else {
                SvfShiftOptions opt(alltdi, allmdi, alltdo, allmdo, JS_SHIFT_IR, endIr);
                boost::shared_ptr<SvfOpSir> p(new SvfOpSir(opt));
                commands.push_back(p);
            }
            //	commands.push_back(op);
        }
        /*
  if (name == "SIR") {
        opt.endState = endIr;
//	SvfOpSir op = SvfOpSir(opt);
//	printIt(op);
//	commands.push_back(op);
        boost::shared_ptr<SvfOpSir> p(new SvfOpSir(opt));
    commands.push_back(p);
  }
*/
    }

};
